from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Type, Union

from sanic import Sanic, __version__
from sanic.exceptions import SanicException
from sanic.log import logger

from sanic_ext.config import Config, add_fallback_config
from sanic_ext.extensions.base import Extension
from sanic_ext.extensions.http.extension import HTTPExtension
from sanic_ext.extensions.injection.extension import InjectionExtension
from sanic_ext.extensions.injection.registry import InjectionRegistry
from sanic_ext.extensions.openapi.extension import OpenAPIExtension

MIN_SUPPORT = (21, 3, 2)


class Extend:
    def __init__(
        self,
        app: Sanic,
        *,
        extensions: Optional[List[Type[Extension]]] = None,
        built_in_extensions: bool = True,
        config: Optional[Union[Config, Dict[str, Any]]] = None,
        **kwargs,
    ) -> None:
        """
        Ingress for instantiating sanic-ext

        :param app: Sanic application
        :type app: Sanic
        """
        if not isinstance(app, Sanic):
            raise SanicException(
                f"Cannot apply SanicExt to {app.__class__.__name__}"
            )

        sanic_version = tuple(map(int, __version__.split(".")))

        if MIN_SUPPORT > sanic_version:
            min_version = ".".join(map(str, MIN_SUPPORT))
            raise SanicException(
                f"SanicExt only works with Sanic v{min_version} and above. "
                f"It looks like you are running {__version__}."
            )

        self.app = app
        self._injection_registry: Optional[InjectionRegistry] = None
        app.ctx.ext = self

        if not isinstance(config, Config):
            config = Config.from_dict(config or {})
        self.config = add_fallback_config(app, config, **kwargs)

        if not extensions:
            extensions = []
        if built_in_extensions:
            extensions.extend(
                [
                    InjectionExtension,
                    OpenAPIExtension,
                    HTTPExtension,
                ]
            )
        init_logs = ["Sanic Extensions:"]
        for extclass in extensions[::-1]:
            extension = extclass(app, self.config)
            extension._startup(self)
            init_logs.append(f"  > {extension.name} {extension.label()}")

        list(map(logger.info, init_logs))

    def injection(
        self,
        type: Type,
        constructor: Optional[Callable[..., Any]] = None,
    ) -> None:
        if not self._injection_registry:
            raise SanicException("Injection extension not enabled")
        self._injection_registry.register(type, constructor)



from ..base import Extension
from .blueprint import blueprint_factory


class OpenAPIExtension(Extension):
    name = "openapi"

    def startup(self, _) -> None:
        if self.app.config.OAS:
            self.bp = blueprint_factory(self.app.config)
            self.app.blueprint(self.bp)

    def label(self):
        if self.app.config.OAS:
            name = f"{self.bp.name}.index"

            if "SERVER_NAME" in self.app.config:
                return f"[{self.app.url_for(name, _external=True)}]"

        return ""



from sanic.log import logger
from sanic.request import Request

from sanic_ext.exceptions import ValidationError


def validate_body(request: Request, args, kwargs, definition):
    model = definition.get(request.content_type, definition.get("*/*", None))
    if model is None:
        raise ValidationError(
            f"Invalid request content-type: {request.content_type}"
        )

    logger.debug(f"Validating {request.path} using {model}")

    try:
        kwargs["body"] = model(**request.json)
    except TypeError as e:
        raise ValidationError(
            f"Invalid request body: {model.__name__}. Error: {e}"
        )


def validate_param(request: Request, args, kwargs, name, schema, location):
    # TODO: MVP

    model = None

    logger.debug(f"Validating {request.args} using {model}")



from inspect import Signature, isawaitable, signature
from typing import Any, Callable, Dict, Optional, Tuple, Type

from sanic import Sanic

from .registry import InjectionRegistry, SignatureRegistry


def add_injection(app: Sanic, injection_registry: InjectionRegistry) -> None:
    signature_registry = _setup_signature_registry(app, injection_registry)

    @app.signal("http.routing.after")
    async def inject_kwargs(request, route, kwargs, **_):
        nonlocal signature_registry

        injections = signature_registry[route.name]

        injected_args = {
            name: await _do_cast(_type, constructor, request, **kwargs)
            for name, (_type, constructor) in injections.items()
        }
        request.match_info.update(injected_args)


async def _do_cast(_type, constructor, request, **kwargs):
    cast = constructor if constructor else _type
    args = [request] if constructor else []
    retval = cast(*args, **kwargs)
    if isawaitable(retval):
        retval = await retval
    return retval


def _setup_signature_registry(
    app: Sanic,
    injection_registry: InjectionRegistry,
) -> SignatureRegistry:
    registry = SignatureRegistry()

    @app.after_server_start
    async def setup_signatures(app, _):
        nonlocal registry

        for route in app.router.routes:
            sig = signature(route.handler)
            injections: Dict[
                str, Tuple[Type, Optional[Callable[..., Any]]]
            ] = {
                param.name: (
                    param.annotation,
                    injection_registry[param.annotation],
                )
                for param in sig.parameters.values()
                if param.annotation != Signature.empty
                and param.annotation in injection_registry._registry
            }
            registry.register(route.name, injections)

    return registry



from sanic_ext.bootstrap import Extend
from sanic_ext.extensions.convenience.decorator import serializer
from sanic_ext.extensions.openapi import openapi

__version__ = "21.9.0a1"
__all__ = [
    "Extend",
    "openapi",
    "serializer",
]



import re
from dataclasses import dataclass
from datetime import timedelta
from typing import FrozenSet, Tuple

from sanic import HTTPResponse, Request, Sanic
from sanic.exceptions import SanicException

WILDCARD_PATTERN = re.compile(r".*")
ORIGIN_HEADER = "access-control-allow-origin"
ALLOW_HEADERS_HEADER = "access-control-allow-headers"
ALLOW_METHODS_HEADER = "access-control-allow-methods"
EXPOSE_HEADER = "access-control-expose-headers"
CREDENTIALS_HEADER = "access-control-allow-credentials"
REQUEST_METHOD_HEADER = "access-control-request-method"
REQUEST_HEADERS_HEADER = "access-control-request-headers"
MAX_AGE_HEADER = "access-control-max-age"
VARY_HEADER = "vary"


@dataclass(frozen=True)
class CORSSettings:
    allow_headers: FrozenSet[str]
    allowed_methods: FrozenSet[str]
    allowed_origins: Tuple[re.Pattern, ...]
    always_send: bool
    automatic_options: bool
    expose_headers: FrozenSet[str]
    max_age: str
    send_wildcard: bool
    supports_credentials: bool


def add_cors(app: Sanic) -> None:
    _setup_cors_settings(app)

    @app.on_response
    async def add_cors_headers(request, response):
        _add_origin_header(request, response)

        if ORIGIN_HEADER not in response.headers:
            return

        _add_expose_header(request, response)
        _add_credentials_header(request, response)
        _add_vary_header(request, response)

        if (
            request.app.ctx.cors.automatic_options
            and request.method == "OPTIONS"
            and request.headers.get(REQUEST_METHOD_HEADER)
        ):
            _add_allow_header(request, response)
            _add_max_age_header(request, response)
            _add_methods_header(request, response)


def _setup_cors_settings(app: Sanic) -> None:
    if app.config.CORS_ORIGINS == "*" and app.config.CORS_SUPPORTS_CREDENTIALS:
        raise SanicException(
            "Cannot use supports_credentials in conjunction with"
            "an origin string of '*'. See: "
            "http://www.w3.org/TR/cors/#resource-requests"
        )

    allow_headers = _get_allow_headers(app)
    allowed_methods = _get_allowed_methods(app)
    allowed_origins = _get_allowed_origins(app)
    expose_headers = _get_expose_headers(app)
    max_age = app.config.CORS_MAX_AGE or ""
    if isinstance(max_age, timedelta):
        max_age = str(int(max_age.total_seconds()))

    app.ctx.cors = CORSSettings(
        allow_headers=allow_headers,
        allowed_methods=allowed_methods,
        allowed_origins=allowed_origins,
        always_send=app.config.CORS_ALWAYS_SEND,
        automatic_options=app.config.CORS_AUTOMATIC_OPTIONS,
        expose_headers=expose_headers,
        max_age=max_age,
        send_wildcard=app.config.CORS_SEND_WILDCARD
        and WILDCARD_PATTERN in allowed_origins,
        supports_credentials=app.config.CORS_SUPPORTS_CREDENTIALS,
    )


def _add_origin_header(request: Request, response: HTTPResponse) -> None:
    request_origin = request.headers.get("origin")
    origin_value = ""

    if request_origin:
        if request.app.ctx.cors.send_wildcard:
            origin_value = "*"
        else:
            for pattern in request.app.ctx.cors.allowed_origins:
                if pattern.match(request_origin):
                    origin_value = request_origin
                    break
    elif request.app.ctx.cors.always_send:
        if WILDCARD_PATTERN in request.app.ctx.cors.allowed_origins:
            origin_value = "*"
        else:
            if (
                isinstance(request.app.config.CORS_ORIGINS, str)
                and "," not in request.app.config.CORS_ORIGINS
            ):
                origin_value = request.app.config.CORS_ORIGINS
            else:
                origin_value = request.app.config.get("SERVER_NAME", "")

    if origin_value:
        response.headers[ORIGIN_HEADER] = origin_value


def _add_expose_header(request: Request, response: HTTPResponse) -> None:
    with_credentials = _is_request_with_credentials(request)
    expose_headers = None
    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials, it is
    # treated as the literal header name "*" without special semantics.
    # Note that the Authorization header can't be wildcarded and always
    # needs to be listed explicitly.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers
    if not with_credentials and "*" in request.app.ctx.cors.expose_headers:
        expose_headers = ["*"]
    elif request.app.ctx.cors.expose_headers:
        expose_headers = request.app.ctx.cors.expose_headers

    if expose_headers:
        response.headers[EXPOSE_HEADER] = ",".join(expose_headers)


def _add_credentials_header(request: Request, response: HTTPResponse) -> None:
    if request.app.ctx.cors.supports_credentials:
        response.headers[CREDENTIALS_HEADER] = "true"


def _add_allow_header(request: Request, response: HTTPResponse) -> None:
    with_credentials = _is_request_with_credentials(request)
    request_headers = set(
        h.strip().lower()
        for h in request.headers.get(REQUEST_HEADERS_HEADER, "").split(",")
    )

    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials,
    # it is treated as the literal header name "*" without special semantics.
    # Note that the Authorization header can't be wildcarded and always needs
    # to be listed explicitly.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers
    if not with_credentials and "*" in request.app.ctx.cors.allow_headers:
        allow_headers = ["*"]
    else:
        allow_headers = request_headers & request.app.ctx.cors.allow_headers

    if allow_headers:
        response.headers[ALLOW_HEADERS_HEADER] = ",".join(allow_headers)


def _add_max_age_header(request: Request, response: HTTPResponse) -> None:
    if request.app.ctx.cors.max_age:
        response.headers[MAX_AGE_HEADER] = request.app.ctx.cors.max_age


def _add_methods_header(request: Request, response: HTTPResponse) -> None:
    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials, it
    # is treated as the literal method name "*" without
    # special semantics.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods
    methods = None
    with_credentials = _is_request_with_credentials(request)
    if not with_credentials and "*" in request.app.ctx.cors.allowed_methods:
        methods = "*"
    elif request.route:
        group = request.app.router.groups.get(request.route.parts)
        if group:
            if request.app.ctx.cors.allowed_methods:
                methods = group.methods & request.app.ctx.cors.allowed_methods
            else:
                methods = group.methods

    if methods:
        response.headers[ALLOW_METHODS_HEADER] = ",".join(methods)


def _add_vary_header(request: Request, response: HTTPResponse) -> None:
    if len(request.app.ctx.cors.allowed_origins) > 1:
        response.headers[VARY_HEADER] = "origin"


def _get_allowed_origins(app: Sanic) -> Tuple[re.Pattern, ...]:
    origins = app.config.CORS_ORIGINS

    if isinstance(origins, str):
        if origins == "*":
            origins = [WILDCARD_PATTERN]
        else:
            origins = origins.split(",")
    elif isinstance(origins, re.Pattern):
        origins = [origins]

    return tuple(
        pattern if isinstance(pattern, re.Pattern) else re.compile(pattern)
        for pattern in origins
    )


def _get_expose_headers(app: Sanic) -> FrozenSet[str]:
    expose_headers = (
        (
            app.config.CORS_EXPOSE_HEADERS
            if isinstance(
                app.config.CORS_EXPOSE_HEADERS, (list, set, frozenset, tuple)
            )
            else app.config.CORS_EXPOSE_HEADERS.split(",")
        )
        if app.config.CORS_EXPOSE_HEADERS
        else tuple()
    )
    return frozenset(header.lower() for header in expose_headers)


def _get_allow_headers(app: Sanic) -> FrozenSet[str]:
    allow_headers = (
        (
            app.config.CORS_ALLOW_HEADERS
            if isinstance(
                app.config.CORS_ALLOW_HEADERS,
                (list, set, frozenset, tuple),
            )
            else app.config.CORS_ALLOW_HEADERS.split(",")
        )
        if app.config.CORS_ALLOW_HEADERS
        else tuple()
    )
    return frozenset(header.lower() for header in allow_headers)


def _get_allowed_methods(app: Sanic) -> FrozenSet[str]:
    allowed_methods = (
        (
            app.config.CORS_METHODS
            if isinstance(
                app.config.CORS_METHODS,
                (list, set, frozenset, tuple),
            )
            else app.config.CORS_METHODS.split(",")
        )
        if app.config.CORS_METHODS
        else tuple()
    )
    return frozenset(method.lower() for method in allowed_methods)


def _is_request_with_credentials(request: Request) -> bool:
    return bool(request.headers.get("authorization") or request.cookies)



"""
This module provides decorators which append
documentation to OperationStore() and components created in the blueprints.

"""
from functools import wraps
from inspect import isclass
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from sanic import Blueprint, Request
from sanic.exceptions import SanicException

from sanic_ext.extensions.openapi.builders import (
    OperationStore,
    SpecificationBuilder,
)
from sanic_ext.extensions.openapi.definitions import (
    Component,
    ExternalDocumentation,
    Parameter,
    RequestBody,
    Response,
    Tag,
)

from .types import Array  # noqa
from .types import Binary  # noqa
from .types import Boolean  # noqa
from .types import Byte  # noqa
from .types import Date  # noqa
from .types import DateTime  # noqa
from .types import Double  # noqa
from .types import Email  # noqa
from .types import Float  # noqa
from .types import Integer  # noqa
from .types import Long  # noqa
from .types import Object  # noqa
from .types import Password  # noqa
from .types import String  # noqa
from .types import Time  # noqa
from .validation import validate_body


def _content_or_component(content):
    if isclass(content):
        spec = SpecificationBuilder()
        if spec._components["schemas"].get(content.__name__):
            content = Component(content)
    return content


def exclude(flag: bool = True, *, bp: Optional[Blueprint] = None):
    if bp:
        for route in bp.routes:
            exclude(flag)(route.handler)
        return

    def inner(func):
        OperationStore()[func].exclude(flag)
        return func

    return inner


def operation(name: str):
    def inner(func):
        OperationStore()[func].name(name)
        return func

    return inner


def summary(text: str):
    def inner(func):
        OperationStore()[func].describe(summary=text)
        return func

    return inner


def description(text: str):
    def inner(func):
        OperationStore()[func].describe(description=text)
        return func

    return inner


def document(url: str, description: str = None):
    def inner(func):
        OperationStore()[func].document(url, description)
        return func

    return inner


def tag(*args: str):
    def inner(func):
        OperationStore()[func].tag(*args)
        return func

    return inner


def deprecated():
    def inner(func):
        OperationStore()[func].deprecate()
        return func

    return inner


def body(
    content: Any,
    validate: Union[
        bool,
        Callable[
            [
                Request,
            ],
            bool,
        ],
    ] = True,
    **kwargs,
):
    def inner(func):
        @wraps(func)
        def handler(request, *handler_args, **handler_kwargs):
            nonlocal kwargs
            nonlocal validate

            if validate:
                if isinstance(validate, bool):
                    validate_body(
                        request, handler_args, handler_kwargs, content
                    )
                else:
                    validate(request, handler_args, handler_kwargs, content)

            return func(request, *handler_args, **handler_kwargs)

        body_content = _content_or_component(content)
        OperationStore()[handler].body(body_content, **kwargs)
        return handler

    return inner


def parameter(name: str, schema: Any, location: str = "query", **kwargs):
    def inner(func):
        OperationStore()[func].parameter(name, schema, location, **kwargs)
        return func

    return inner


def response(status, content: Any = None, description: str = None, **kwargs):
    def inner(func):
        OperationStore()[func].response(status, content, description, **kwargs)
        return func

    return inner


def secured(*args, **kwargs):
    raise NotImplementedError(
        "SecuritySchemas are not yet implemented in sanic-openapi 0.6.3, "
        "hopefully they should be ready for the next release."
    )

    def inner(func):
        OperationStore()[func].secured(*args, **kwargs)
        return func

    return inner


Model = TypeVar("Model")


def component(
    model: Optional[Model] = None,
    *,
    name: Optional[str] = None,
    field: Optional[str] = None,
):
    def wrap(m):
        return component(m, name=name, field=field)

    if not model:
        return wrap

    params = {}
    if name:
        params["name"] = name
    if field:
        params["field"] = field
    Component(model, **params)
    return model


def definition(
    *,
    exclude: Optional[bool] = None,
    operation: Optional[str] = None,
    summary: Optional[str] = None,
    description: Optional[str] = None,
    document: Optional[Union[str, ExternalDocumentation]] = None,
    tag: Optional[Union[Union[str, Tag], Sequence[Union[str, Tag]]]] = None,
    deprecated: bool = False,
    body: Optional[Union[Dict[str, Any], RequestBody, Any]] = None,
    parameter: Optional[
        Union[
            Union[Dict[str, Any], Parameter, str],
            List[Union[Dict[str, Any], Parameter, str]],
        ]
    ] = None,
    response: Optional[
        Union[
            Union[Dict[str, Any], Response, Any],
            List[Union[Dict[str, Any], Response, Any]],
        ]
    ] = None,
):
    def inner(func):
        glbl = globals()

        if exclude is not None:
            func = glbl["exclude"](exclude)(func)

        if operation:
            func = glbl["operation"](operation)(func)

        if summary:
            func = glbl["summary"](summary)(func)

        if description:
            func = glbl["description"](description)(func)

        if document:
            kwargs = {}
            if isinstance(document, str):
                kwargs["url"] = document
            else:
                kwargs["url"] = document.fields["url"]
                kwargs["description"] = document.fields["description"]

            func = glbl["document"](**kwargs)(func)

        if tag:
            taglist = []
            op = (
                "extend"
                if isinstance(tag, (list, tuple, set, frozenset))
                else "append"
            )

            getattr(taglist, op)(tag)
            func = glbl["tag"](
                *[
                    tag.fields["name"] if isinstance(tag, Tag) else tag
                    for tag in taglist
                ]
            )(func)

        if deprecated:
            func = glbl["deprecated"]()(func)

        if body:
            kwargs = {}
            content = body
            if isinstance(content, RequestBody):
                kwargs = content.fields
            elif isinstance(content, dict):
                if "content" in content:
                    kwargs = content
                else:
                    kwargs["content"] = content
            else:
                content = _content_or_component(content)
                kwargs["content"] = content
            func = glbl["body"](**kwargs)(func)

        if parameter:
            paramlist = []
            op = (
                "extend"
                if isinstance(parameter, (list, tuple, set, frozenset))
                else "append"
            )
            getattr(paramlist, op)(parameter)

            for param in paramlist:
                kwargs = {}
                if isinstance(param, Parameter):
                    kwargs = param.fields
                elif isinstance(param, dict) and "name" in param:
                    kwargs = param
                elif isinstance(param, str):
                    kwargs["name"] = param
                else:
                    raise SanicException(
                        "parameter must be a Parameter instance, a string, or "
                        "a dictionary containing at least 'name'."
                    )

                if "schema" not in kwargs:
                    kwargs["schema"] = str

                func = glbl["parameter"](**kwargs)(func)

        if response:
            resplist = []
            op = (
                "extend"
                if isinstance(response, (list, tuple, set, frozenset))
                else "append"
            )
            getattr(resplist, op)(response)

            for resp in resplist:
                kwargs = {}
                if isinstance(resp, Response):
                    kwargs = resp.fields
                elif isinstance(resp, dict):
                    if "content" in resp:
                        kwargs = resp
                    else:
                        kwargs["content"] = resp
                else:
                    kwargs["content"] = resp

                if "status" not in kwargs:
                    kwargs["status"] = 200

                func = glbl["response"](**kwargs)(func)

        return func

    return inner



"""
Sanic
"""
import codecs
import os
import re

from setuptools import find_packages, setup

install_requires = [
    "pyyaml>=3.0.0",
]

dev_requires = ["black==19.3b0", "flake8==3.7.7", "isort==4.3.19"]

test_requires = [
    "sanic_testing==0.7.0b2",
    "coverage",
    "pytest",
    "pytest-cov",
    "tox",
]

project_root = os.path.dirname(os.path.abspath(__file__))

with codecs.open(
    os.path.join(project_root, "sanic_ext", "__init__.py"), "r", "latin1"
) as fp:
    try:
        version = re.findall(
            r"^__version__ = \"([^']+)\"\r?$", fp.read(), re.M
        )[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

with open(os.path.join(project_root, "README.md"), "r") as f:
    long_description = f.read()

setup(
    name="sanic-ext",
    version=version,
    url="http://github.com/sanic-org/sanic-ext/",
    license="MIT",
    author="Sanic Community",
    description="Extend your Sanic installation with some core functionality.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_data={"sanic_ext": ["extensions/openapi/ui/*"]},
    platforms="any",
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires + test_requires,
        "test": test_requires,
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
